@@ -301,14 +301,16 @@ class Datacube(object):
         group_by = query_group_by(**query)
         grouped = self.group_datasets(observations, group_by)
 
-        measurements = self.index.products.get_by_name(product).lookup_measurements(measurements)
-        measurements = set_resampling_method(measurements, resampling)
+        datacube_product = self.index.products.get_by_name(product)
+        measurement_dicts = set_resampling_method(datacube_product.lookup_measurements(measurements),
+                                                  resampling)
 
-        result = self.load_data(grouped, geobox, list(measurements.values()),
+        result = self.load_data(grouped, geobox, list(measurement_dicts.values()),
                                 fuse_func=fuse_func,
                                 dask_chunks=dask_chunks,
                                 use_threads=use_threads)
-        return result
+
+        return apply_aliases(result, datacube_product, measurements)
 
     def product_observations(self, **kwargs):
         warnings.warn("product_observations() has been renamed to find_datasets() and will eventually be removed",
@@ -439,17 +441,8 @@ class Datacube(object):
 
         for measurement in measurements:
             data = results.pop(0)
-
-            attrs = {
-                'nodata': measurement.get('nodata'),
-                'units': measurement.get('units', '1'),
-                'crs': geobox.crs
-            }
-            if 'flags_definition' in measurement:
-                attrs['flags_definition'] = measurement['flags_definition']
-            if 'spectral_definition' in measurement:
-                attrs['spectral_definition'] = measurement['spectral_definition']
-
+            attrs = measurement.dataarray_attrs()
+            attrs['crs'] = geobox.crs
             dims = tuple(coords.keys()) + tuple(geobox.dimensions)
             result[measurement['name']] = (dims, data, attrs)
 
@@ -578,6 +571,17 @@ class Datacube(object):
         self.close()
 
 
+def apply_aliases(data, product, measurements):
+    """
+    If measurements are referred to by their aliases,
+    rename data arrays to reflect that.
+    """
+    if measurements is None:
+        return data
+
+    return data.rename({product.canonical_measurement(provided_name): provided_name
+                        for provided_name in measurements})
+
 def output_geobox(like=None, output_crs=None, resolution=None, align=None,
                   grid_spec=None, datasets=None, **query):
     """ Configure output geobox from user provided output specs. """
