@@ -14,7 +14,7 @@ from uuid import UUID
 from affine import Affine
 
 from datacube.compat import urlparse
-from datacube.utils import geometry
+from datacube.utils import geometry, without_lineage_sources
 from datacube.utils import parse_time, cached_property, uri_to_local_path, intersects, schema_validated, DocReader
 from datacube.utils.geometry import (CRS as _CRS,
                                      GeoBox as _GeoBox,
@@ -299,9 +299,10 @@ class Dataset(object):
         return hash(self.id)
 
     def __str__(self):
+        str_loc = 'not available' if not self.uris else self.uris[0]
         return "Dataset <id={id} type={type} location={loc}>".format(id=self.id,
                                                                      type=self.type.name,
-                                                                     loc=self.local_path)
+                                                                     loc=str_loc)
 
     def __repr__(self):
         return self.__str__()
@@ -310,22 +311,42 @@ class Dataset(object):
     def metadata(self):
         return self.metadata_type.dataset_reader(self.metadata_doc)
 
+    def metadata_doc_without_lineage(self):
+        """ Return metadata document without nested lineage datasets
+        """
+        return without_lineage_sources(self.metadata_doc, self.metadata_type)
 
-class Measurement(object):
-    REQUIRED_KEYS = ['name', 'dtype', 'nodata', 'units']
-    OPTIONAL_KEYS = ['aliases', 'spectral_definition', 'flags_definition']
+
+class Measurement(dict):
+    REQUIRED_KEYS = ('name', 'dtype', 'nodata', 'units')
+    OPTIONAL_KEYS = ('aliases', 'spectral_definition', 'flags_definition')
+    FILTER_ATTR_KEYS = ('name', 'dtype', 'aliases')
 
     def __init__(self, **measurement_dict):
-        for key in self.REQUIRED_KEYS:
-            setattr(self, key, measurement_dict[key])
+        missing_keys = set(self.REQUIRED_KEYS) - set(measurement_dict)
+        if missing_keys:
+            raise ValueError("Measurement required keys missing: {}".format(missing_keys))
 
-        for key in self.OPTIONAL_KEYS:
-            if key in measurement_dict:
-                setattr(self, key, measurement_dict[key])
+        measurement_data = {key: value for key, value in measurement_dict.items()
+                            if key in self.REQUIRED_KEYS + self.OPTIONAL_KEYS}
+
+        super().__init__(measurement_data)
+
+    def __getattr__(self, key):
+        """ Allow access to items as attributes. """
+        return self[key]
 
     def __repr__(self):
-        attrs = ", ".join("{}={}".format(key, value) for key, value in vars(self).items())
-        return "Measurement({})".format(attrs)
+        return "Measurement({})".format(super(Measurement, self).__repr__())
+
+    def copy(self):
+        """Required as the super class `dict` method returns a `dict`
+           and does not preserve Measurement class"""
+        return Measurement(**self)
+
+    def dataarray_attrs(self):
+        """This returns attributes filtered for display in a dataarray."""
+        return {key: value for key, value in self.items() if key not in self.FILTER_ATTR_KEYS}
 
 
 @schema_validated(SCHEMA_PATH / 'metadata-type-schema.yaml')
@@ -417,7 +438,7 @@ class DatasetType(object):
 
         :type: dict[str, dict]
         """
-        return OrderedDict((m['name'], m) for m in self.definition.get('measurements', []))
+        return OrderedDict((m['name'], Measurement(**m)) for m in self.definition.get('measurements', []))
 
     @property
     def dimensions(self):
@@ -463,6 +484,14 @@ class DatasetType(object):
 
         return GridSpec(crs=crs, tile_size=tile_size, resolution=resolution, origin=origin)
 
+    def canonical_measurement(self, measurement):
+        for m in self.measurements:
+            if measurement == m:
+                return measurement
+            elif measurement in self.measurements[m].get('aliases', []):
+                return m
+        raise KeyError(measurement)
+
     def lookup_measurements(self, measurements=None):
         """
         Find measurements by name
@@ -473,7 +502,8 @@ class DatasetType(object):
         my_measurements = self.measurements
         if measurements is None:
             return my_measurements
-        return OrderedDict((measurement, my_measurements[measurement]) for measurement in measurements)
+        canonical = [self.canonical_measurement(measurement) for measurement in measurements]
+        return OrderedDict((measurement, my_measurements[measurement]) for measurement in canonical)
 
     def dataset_reader(self, dataset_doc):
         return self.metadata_type.dataset_reader(dataset_doc)
@@ -710,3 +740,13 @@ class GridSpec(object):
 
     def __repr__(self):
         return self.__str__()
+
+
+def metadata_from_doc(doc):
+    """Construct MetadataType that is not tied to any particular db index. This is
+    useful when there is a need to interpret dataset metadata documents
+    according to metadata spec.
+    """
+    from .fields import get_dataset_fields
+    MetadataType.validate(doc)
+    return MetadataType(doc, get_dataset_fields(doc))
