@@ -5,35 +5,27 @@ import datetime
 import logging
 import sys
 from collections import OrderedDict
-from pathlib import Path
 
 import click
 import yaml
 import yaml.resolver
 from click import echo
-import json
 
 from datacube.index.index import Index
 from datacube.index.exceptions import MissingRecordError
 from datacube.model import Dataset
+from datacube.index.hl import Doc2Dataset, check_dataset_consistent
 from datacube.ui import click as ui
 from datacube.ui.click import cli
-from datacube.ui.common import get_metadata_path
-from datacube.utils import read_documents, changes, InvalidDocException, without_lineage_sources
+from datacube.ui.common import ui_path_doc_stream
+from datacube.utils import changes
 from datacube.utils.serialise import SafeDatacubeDumper
 
-try:
-    from typing import Iterable
-except ImportError:
-    pass
+from typing import Iterable
 
 _LOG = logging.getLogger('datacube-dataset')
 
 
-class BadMatch(Exception):
-    pass
-
-
 def report_old_options(mapping):
     def maybe_remap(s):
         if s in mapping:
@@ -49,150 +41,56 @@ def dataset_cmd():
     pass
 
 
-def find_matching_product(rules, doc):
-    """:rtype: datacube.model.DatasetType"""
-    matched = [rule for rule in rules if changes.contains(doc, rule['metadata'])]
-    if not matched:
-        # provide user with information about the failure
-        if len(rules) == 0:
-            raise BadMatch('No rules provided.')
-        elif len(rules) == 1:
-            metadata = rules[0]['metadata']
-            relevant_doc = {k: v for k, v in doc.items() if k in metadata}
-            raise BadMatch('Dataset metadata did not match product rules.'
-                           '\nDataset metadata:\n %s\n'
-                           '\nProduct metadata:\n %s\n'
-                           % (json.dumps(relevant_doc, indent=4),
-                              json.dumps(metadata, indent=4)))
-        else:
-            raise BadMatch('No matching Product found for %s' % json.dumps(doc, indent=4))
-    if len(matched) > 1:
-        raise BadMatch('Too many matching Products found for %s. Matched %s.' % (
-            doc.get('id', 'unidentified'), matched))
-    return matched[0]['type']
+def dataset_stream(doc_stream, ds_resolve):
+    """ Convert a stream `(uri, doc)` pairs into a stream of resolved datasets
 
-
-def check_dataset_consistent(dataset):
+        skips failures with logging
     """
-    :type dataset: datacube.model.Dataset
-    :return: (Is consistent, [error message|None])
-    :rtype: (bool, str or None)
-    """
-    # It the type expects measurements, ensure our dataset contains them all.
-    if not set(dataset.type.measurements.keys()).issubset(dataset.measurements.keys()):
-        return False, "measurement fields don't match type specification"
+    for uri, ds in doc_stream:
+        dataset, err = ds_resolve(ds, uri)
 
-    return True, None
-
-
-def create_dataset(dataset_doc, uri, rules, skip_lineage=False):
-    """
-    :rtype datacube.model.Dataset:
-    """
-    dataset_type = find_matching_product(rules, dataset_doc)
-    if skip_lineage:
-        sources = {}
-        dataset_doc = without_lineage_sources(dataset_doc, dataset_type)
-    else:
-        sources = {cls: create_dataset(source_doc, None, rules)
-                   for cls, source_doc in dataset_type.dataset_reader(dataset_doc).sources.items()}
-    return Dataset(dataset_type, dataset_doc, uris=[uri] if uri else None, sources=sources)
-
-
-def load_rules_from_types(index, type_names=None):
-    types = []
-    if type_names:
-        for name in type_names:
-            type_ = index.products.get_by_name(name)
-            if not type_:
-                _LOG.error('DatasetType %s does not exists', name)
-                return None
-            types.append(type_)
-    else:
-        types += index.products.get_all()
-
-    rules = [{'type': type_, 'metadata': type_.metadata_doc} for type_ in types]
-    return rules
-
-
-def load_datasets(datasets, rules, skip_lineage=False):
-    for dataset_path in datasets:
-        metadata_path = get_metadata_path(Path(dataset_path))
-        if not metadata_path or not metadata_path.exists():
-            _LOG.error('No supported metadata docs found for dataset %s', dataset_path)
+        if dataset is None:
+            _LOG.error('%s', str(err))
             continue
 
-        try:
-            for uri, metadata_doc in read_documents(metadata_path, uri=True):
-                try:
-                    dataset = create_dataset(metadata_doc, uri, rules, skip_lineage=skip_lineage)
-                except BadMatch as e:
-                    _LOG.error('Unable to create Dataset for %s: %s', uri, e)
-                    continue
-
-                is_consistent, reason = check_dataset_consistent(dataset)
-                if not is_consistent:
-                    _LOG.error("Dataset %s inconsistency: %s", dataset.id, reason)
-                    continue
-
-                yield dataset
-        except InvalidDocException:
-            _LOG.error("Failed reading documents from %s", metadata_path)
-            continue
+        yield dataset
 
 
-def load_datasets_for_update(datasets, index):
-    """Load datasets from disk, associate to a product by looking up existing
-    dataset in the index. Datasets not in the database will be logged.
+def load_datasets_for_update(doc_stream, index):
+    """Consume stream of dataset documents, associate each to a product by looking
+    up existing dataset in the index. Datasets not in the database will be
+    logged.
 
     Doesn't load lineage information
 
     Generates tuples in the form (new_dataset, existing_dataset)
     """
-    def mk_dataset(metadata_doc, uri):
-        uuid = metadata_doc.get('id', None)
+    def mk_dataset(ds, uri):
+        uuid = ds.id
 
         if uuid is None:
             return None, None, "Metadata document it missing id field"
 
         existing = index.datasets.get(uuid)
         if existing is None:
-            return None, None, "No such dataset the database: {}".format(uuid)
+            return None, None, "No such dataset in the database: {}".format(uuid)
 
         return Dataset(existing.type,
-                       without_lineage_sources(metadata_doc, existing.type, inplace=True),
+                       ds.doc_without_lineage_sources,
                        uris=[uri]), existing, None
 
-    for dataset_path in datasets:
-        metadata_path = get_metadata_path(Path(dataset_path))
-        if not metadata_path or not metadata_path.exists():
-            _LOG.error('No supported metadata docs found for dataset %s', dataset_path)
-            continue
+    for uri, doc in doc_stream:
+        dataset, existing, error_msg = mk_dataset(doc, uri)
 
-        try:
-            for uri, metadata_doc in read_documents(metadata_path, uri=True):
-                dataset, existing, error_msg = mk_dataset(metadata_doc, uri)
-
-                if dataset is None:
-                    _LOG.error("Failure while processing: %s\n" +
-                               " > Reason: %s", metadata_path, error_msg)
-                else:
-                    is_consistent, reason = check_dataset_consistent(dataset)
-                    if is_consistent:
-                        yield dataset, existing
-                    else:
-                        _LOG.error("Dataset %s inconsistency: %s", dataset.id, reason)
-
-        except InvalidDocException:
-            _LOG.error("Failed reading documents from %s", metadata_path)
-            continue
-
-
-def parse_match_rules_options(index, dtype, auto_match):
-    if auto_match is True:
-        _LOG.warning("--auto-match option is deprecated, update your scripts, behaviour is the same without it")
-
-    return load_rules_from_types(index, dtype)
+        if dataset is None:
+            _LOG.error("Failure while processing: %s\n" +
+                       " > Reason: %s", uri, error_msg)
+        else:
+            is_consistent, reason = check_dataset_consistent(dataset)
+            if is_consistent:
+                yield dataset, existing
+            else:
+                _LOG.error("Dataset %s inconsistency: %s", dataset.id, reason)
 
 
 @dataset_cmd.command('add',
@@ -205,23 +103,37 @@ def parse_match_rules_options(index, dtype, auto_match):
               help=('Only match against products specified with this option, '
                     'you can supply several by repeating this option with a new product name'),
               multiple=True)
+@click.option('--exclude-product', '-x', 'exclude_product_names',
+              help=('Attempt to match to all products in the DB except for products '
+                    'specified with this option, '
+                    'you can supply several by repeating this option with a new product name'),
+              multiple=True)
 @click.option('--auto-match', '-a', help="Deprecated don't use it, it's a no-op",
               is_flag=True, default=False)
-@click.option('--sources-policy', type=click.Choice(['verify', 'ensure', 'skip']), default='verify',
-              help="""'verify' - verify source datasets' metadata (default)
-'ensure' - add source dataset if it doesn't exist
-'skip' - dont add the derived dataset if source dataset doesn't exist""")
+@click.option('--auto-add-lineage/--no-auto-add-lineage', is_flag=True, default=True,
+              help=('Default behaviour is to automatically add lineage datasets if they are missing from the database, '
+                    'but this can be disabled if lineage is expected to be present in the DB, '
+                    'in this case add will abort when encountering missing lineage dataset'))
+@click.option('--verify-lineage/--no-verify-lineage', is_flag=True, default=True,
+              help=('Lineage referenced in the metadata document should be the same as in DB, '
+                    'default behaviour is to skip those top-level datasets that have lineage data '
+                    'different from the version in the DB. This option allows omitting verification step.'))
 @click.option('--dry-run', help='Check if everything is ok', is_flag=True, default=False)
 @click.option('--ignore-lineage',
-              help="Don't add lineage data to the database",
+              help="Pretend that there is no lineage data in the datasets being indexed",
               is_flag=True, default=False)
 @click.option('--confirm-ignore-lineage',
-              help="Don't add lineage data to the database, without confirmation",
+              help="Pretend that there is no lineage data in the datasets being indexed, without confirmation",
               is_flag=True, default=False)
 @click.argument('dataset-paths',
                 type=click.Path(exists=True, readable=True, writable=False), nargs=-1)
 @ui.pass_index()
-def index_cmd(index, product_names, auto_match, sources_policy, dry_run,
+def index_cmd(index, product_names,
+              exclude_product_names,
+              auto_match,
+              auto_add_lineage,
+              verify_lineage,
+              dry_run,
               ignore_lineage,
               confirm_ignore_lineage,
               dataset_paths):
@@ -238,26 +150,42 @@ def index_cmd(index, product_names, auto_match, sources_policy, dry_run,
 
         confirm_ignore_lineage = True
 
-    rules = parse_match_rules_options(index, product_names, auto_match)
-    if rules is None:
-        return
+    if auto_match is True:
+        _LOG.warning("--auto-match option is deprecated, update your scripts, behaviour is the same without it")
+
+    try:
+        ds_resolve = Doc2Dataset(index,
+                                 product_names,
+                                 exclude_products=exclude_product_names,
+                                 skip_lineage=confirm_ignore_lineage,
+                                 fail_on_missing_lineage=not auto_add_lineage,
+                                 verify_lineage=verify_lineage)
+    except ValueError as e:
+        _LOG.error(e)
+        sys.exit(2)
+
+    def run_it(dataset_paths):
+        doc_stream = ui_path_doc_stream(dataset_paths, logger=_LOG, uri=True)
+        dss = dataset_stream(doc_stream, ds_resolve)
+        index_datasets(dss,
+                       index,
+                       auto_add_lineage=auto_add_lineage,
+                       dry_run=dry_run)
 
     # If outputting directly to terminal, show a progress bar.
     if sys.stdout.isatty():
-        with click.progressbar(dataset_paths, label='Indexing datasets') as dataset_path_iter:
-            index_dataset_paths(sources_policy, dry_run, index, rules, dataset_path_iter,
-                                skip_lineage=confirm_ignore_lineage)
+        with click.progressbar(dataset_paths, label='Indexing datasets') as pp:
+            run_it(pp)
     else:
-        index_dataset_paths(sources_policy, dry_run, index, rules, dataset_paths,
-                            skip_lineage=confirm_ignore_lineage)
+        run_it(dataset_paths)
 
 
-def index_dataset_paths(sources_policy, dry_run, index, rules, dataset_paths, skip_lineage=False):
-    for dataset in load_datasets(dataset_paths, rules, skip_lineage=skip_lineage):
+def index_datasets(dss, index, auto_add_lineage, dry_run):
+    for dataset in dss:
         _LOG.info('Matched %s', dataset)
         if not dry_run:
             try:
-                index.datasets.add(dataset, sources_policy=sources_policy)
+                index.datasets.add(dataset, with_lineage=auto_add_lineage)
             except (ValueError, MissingRecordError) as e:
                 _LOG.error('Failed to add dataset %s: %s', dataset.local_uri, e)
 
@@ -282,10 +210,10 @@ def parse_update_rules(keys_that_can_change):
 'archive' - mark as archived
 'forget' - remove from the index
 ''')
-@click.argument('datasets',
+@click.argument('dataset-paths',
                 type=click.Path(exists=True, readable=True, writable=False), nargs=-1)
 @ui.pass_index()
-def update_cmd(index, keys_that_can_change, dry_run, location_policy, datasets):
+def update_cmd(index, keys_that_can_change, dry_run, location_policy, dataset_paths):
 
     def loc_action(action, new_ds, existing_ds, action_name):
         if len(existing_ds.uris) == 0:
@@ -323,7 +251,9 @@ def update_cmd(index, keys_that_can_change, dry_run, location_policy, datasets):
     updates_allowed = parse_update_rules(keys_that_can_change)
 
     success, fail = 0, 0
-    for dataset, existing_ds in load_datasets_for_update(datasets, index):
+
+    for dataset, existing_ds in load_datasets_for_update(
+            ui_path_doc_stream(dataset_paths, logger=_LOG, uri=True), index):
         _LOG.info('Matched %s', dataset)
 
         if location_policy != 'keep':
