@@ -112,7 +112,7 @@ def _calc_offsets2(off, scale, src_size, dst_size):
         return _calc_offsets_impl(off, scale, src_size, dst_size)
 
 
-def _read_decimated(array_transform, src, dest_shape):
+def _read_native(array_transform, src, dest_shape):
     dy_dx = (array_transform.f, array_transform.c)
     sy_sx = (array_transform.e, array_transform.a)
     read, write, read_shape, write_shape = zip(*map(_calc_offsets2, dy_dx, sy_sx, src.shape, dest_shape))
@@ -145,21 +145,22 @@ def read_from_source(source, dest, dst_transform, dst_nodata, dst_projection, re
     """
     with source.open() as src:
         array_transform = ~src.transform * dst_transform
-        # if the CRS is the same use decimated reads if possible (NN or 1:1 scaling)
-        can_use_decimated_read = (src.crs == dst_projection and
-                                  _no_scale(array_transform) and
-                                  (resampling == Resampling.nearest or _no_fractional_translate(array_transform)))
-        if can_use_decimated_read:
+        # if the CRS is the same use native reads if possible (NN or 1:1 scaling)
+        can_use_native_read = (src.crs == dst_projection and
+                               _no_scale(array_transform) and
+                               (resampling == Resampling.nearest or _no_fractional_translate(array_transform)))
+        if can_use_native_read:
             dest.fill(dst_nodata)
             try:
-                tmp, offset, _ = _read_decimated(array_transform, src, dest.shape)
+                tmp, offset, _ = _read_native(array_transform, src, dest.shape)
                 if tmp is None:
                     return
             except ValueError:
                 _LOG.debug('Failed Read: %s', src, exc_info=1)
                 return
             dest = dest[offset[0]:offset[0] + tmp.shape[0], offset[1]:offset[1] + tmp.shape[1]]
-            numpy.copyto(dest, tmp, where=(tmp != src.nodata))
+            where_valid_data = (tmp != src.nodata) if not numpy.isnan(src.nodata) else ~numpy.isnan(tmp)
+            numpy.copyto(dest, tmp, where=where_valid_data)
         else:
             if dest.dtype == numpy.dtype('int8'):
                 dest = dest.view(dtype='uint8')
@@ -192,7 +193,8 @@ def reproject_and_fuse(datasources, destination, dst_transform, dst_projection,
         :type dest: numpy.ndarray
         :type src: numpy.ndarray
         """
-        numpy.copyto(dest, src, where=(dest == dst_nodata))
+        where_nodata = (dest == dst_nodata) if not numpy.isnan(dst_nodata) else numpy.isnan(dest)
+        numpy.copyto(dest, src, where=where_nodata)
 
     fuse_func = fuse_func or copyto_fuser
 
